console = require 'console'

star = {}
stars = {}
dynamicStars = {}

viewport = {}
viewport.x = 0
viewport.y = 0
viewport.w = love.graphics.getWidth()
viewport.h = love.graphics.getHeight()

world = {}
world.x = 0
world.y = 0
world.w = 3000
world.h = 3000

ship = {}
ship.img = {}
ship.x = 100
ship.y = 100
ship.xVel = 0
ship.yVel = 0
ship.speed = 0
ship.maxSpeed = 400
ship.accel = ship.maxSpeed * 0.4
ship.deccel = ship.maxSpeed * 0.5
ship.angle = 0
ship.angleSpeed = math.pi

function love.load()
    console.load()
     
    stars = initStarField(2000)
    dynamicStars = initDynamicStars(40)
    
    ship.img = love.graphics.newImage("8B.png")
end

function love.draw()
    -- transform based on viewport coords
    love.graphics.push()
    love.graphics.translate(-viewport.x, -viewport.y)
    
    -- draw borders
    drawBorders()
    
    -- draw star field
    love.graphics.setColor(255, 255, 255)
    love.graphics.setPointStyle("rough")
    for i = 1, #stars do
        local star = stars[i]
        if isStarVisible(star) then
            love.graphics.setPointSize(star.size)
            love.graphics.point(star.x, star.y)
        end
    end
    
    -- draw dynamic stars
    --love.graphics.setColor(0, 155, 200)
    love.graphics.setColor(0, 255, 255)
    love.graphics.setPointStyle("rough")
    for i = 1, #dynamicStars do
        local star = dynamicStars[i]
        if isStarVisible(star) then
            love.graphics.setPointSize(star.size)
            love.graphics.point(star.x, star.y)
        end
    end
    
    -- draw ship
    love.graphics.setColor(255, 255, 255)
    love.graphics.draw(ship.img, ship.x, ship.y, ship.angle, 1, 1, ship.img:getWidth() / 2, ship.img:getHeight() / 2)
    
    -- return to normal view and draw dat console
    love.graphics.pop()
    console.draw()
end

function love.update(dt)
    -- only process updates if console isn't active
    if console.active == false then
        if love.keyboard.isDown("w") then
            ship.speed = ship.speed + (ship.accel * dt)
            if ship.speed >= ship.maxSpeed then ship.speed = ship.maxSpeed end
        end
        if love.keyboard.isDown("s") then
            ship.speed = ship.speed - (ship.deccel * dt)
            if ship.speed <= 0 then ship.speed = 0 end
        end
        
        -- calculate ship rotation angle
        --ship.angle = math.atan2(love.mouse.getY() - ship.y, love.mouse.getX() - ship.x)
        if love.keyboard.isDown("a") then
           ship.angle = ship.angle - ship.angleSpeed * dt
        end
        if love.keyboard.isDown("d") then
           ship.angle = ship.angle + ship.angleSpeed * dt
        end
        
        -- ship velocity
        ship.xVel = math.cos(ship.angle) * ship.speed
        ship.yVel = math.sin(ship.angle) * ship.speed
        
        -- update ship position
        ship.x = ship.x + ship.xVel * dt
        ship.y = ship.y + ship.yVel * dt
        
        -- adjust viewport to center on ship
        updateViewport()
        
        -- update dynamic star positions based on how fast the ship is moving
        updateDynamicStars(ship.xVel * dt, ship.yVel * dt)
    end
end

function love.keypressed(key, unicode)
    console.keypressed(key, unicode)
    
    if key == 'escape' then
        love.event.push('quit') -- Quit the game.
    end 
end

function star:new(minSize, maxSize, x, y, w, h)
    local t = {}
    
    t.x = math.random(x, w)
    t.y = math.random(y, h)
    t.size = math.random(minSize, maxSize)
    
    return t
end

function initStarField(maxStars)
    local minSize, maxSize = 1, 6
    local starField = {}
    
    for i = 1, maxStars do
        starField[#starField + 1] = star:new(minSize, maxSize, world.x, world.y, world.w, world.h)
    end
    
    return starField
end

function isStarVisible(star)
    if star.x >= viewport.x and star.x <= (viewport.x + viewport.w) then
        if star.y >= viewport.y and star.y <= (viewport.y + viewport.h) then
            return true
        else
            return false
        end
    else
        return false
    end
end

function updateViewport()
    viewport.x = math.floor(ship.x - (viewport.w / 2))
    viewport.y = math.floor(ship.y - (viewport.h / 2))
    
    -- ensure within bounds of world
    if viewport.x < world.x then viewport.x = world.x end
    if (viewport.x + viewport.w) > world.w then viewport.x = (world.w - viewport.w) end
    if viewport.y < world.y then viewport.y = world.y end
    if (viewport.y + viewport.h) > world.h then viewport.y = (world.h - viewport.h) end
end

function drawBorders()
    -- draw world "edges"
    if viewport.x <= world.x then
        love.graphics.setColor({255, 0, 255})
        love.graphics.setLineWidth(4)
        love.graphics.line(viewport.x, viewport.y, viewport.x, viewport.y + viewport.h)
    end
    if (viewport.x + viewport.w) >= world.w then
        love.graphics.setColor({255, 0, 255})
        love.graphics.setLineWidth(4)
        love.graphics.line(viewport.x + viewport.w, viewport.y, viewport.x + viewport.w, viewport.y + viewport.h)
    end
    if viewport.y <= world.y then
        love.graphics.setColor({255, 0, 255})
        love.graphics.setLineWidth(4)
        love.graphics.line(viewport.x, viewport.y, viewport.x + viewport.w, viewport.y)
    end
    if (viewport.y + viewport.h) >= world.h then
        love.graphics.setColor({255, 0, 255})
        love.graphics.setLineWidth(4)
        love.graphics.line(viewport.x, viewport.y + viewport.h, viewport.x + viewport.w, viewport.y + viewport.h)
    end
end

function initDynamicStars(maxStars)
    local minSize, maxSize = 4, 8
    local stars = {}
    
    for i = 1, maxStars do
        stars[#stars + 1] = star:new(minSize, maxSize, viewport.x, viewport.y, viewport.w, viewport.h)
    end

    return stars
end

function updateDynamicStars(dx, dy)
    if dx == 0 and dy == 0 then return end
    
    for i = 1, #dynamicStars do
        local star = dynamicStars[i]
        
        -- update star position
        star.x = star.x + math.ceil((-dx  * 0.25 * star.size))
        star.y = star.y + math.ceil((-dy  * 0.25 * star.size))
        
        -- Free up stars that have gone offscreen and create new ones
        if not isStarVisible(star) then
            -- depending on dx and dy, calculate which edges the star should spawn at
            -- we do this by figuring out if x or y change was positive or negative; this
            -- gets us camera movement so that stars spawn on the edge of where the camera
            -- is moving towards. also, lua is cool and we don't have to worry bout div/0
            -- it'll just eval to -nan
            local edges = {}
            if (dx / dx) >= 1 then edges[#edges + 1] = "right" end
            if (dx / dx) <= 1 then edges[#edges + 1] = "left" end
            if (dy / dy) >= 1 then edges[#edges + 1] = "bottom" end
            if (dy / dy) <= 1 then edges[#edges + 1] = "top" end
            
            -- We now know which edge(s) the star will spawn, so we can randomly spawn 'em
            local edge = math.random(1, #edges)
            if edges[edge] == "right" then
                star.x = viewport.x + viewport.w
                star.y = math.random(viewport.y, viewport.y + viewport.h)
            elseif edges[edge] == "left" then
                star.x = viewport.x
                star.y = math.random(viewport.y, viewport.y + viewport.h)            
            elseif edges[edge] == "bottom" then
                star.x = math.random(viewport.x, viewport.x + viewport.w)
                star.y = viewport.y            
            elseif edges[edge] == "top" then
                star.x = math.random(viewport.x, viewport.x + viewport.w)
                star.y = viewport.y + viewport.h
            end
        end
    end
end